from concurrent.futures import ProcessPoolExecutor as Pool
from math import ceil, log10
from multiprocessing import cpu_count
from os import mkdir
from os.path import isdir
from time import time

from numba import njit
import numpy as np

from sphractal.utils import calcDist, oppositeInnerAtoms
# from sphractal.utils import annotate


@njit(fastmath=True, cache=True)
def getNearFarCoord(scanBoxIdx, boxLen, lowBound, atomCoord, bufferDist=5.0):
    """Find the nearest and furthest point of a given box from a given atom."""
    scanBoxMax = lowBound - bufferDist + (scanBoxIdx+1)*boxLen
    scanBoxMin = scanBoxMax - boxLen
    if atomCoord < scanBoxMin:
        scanBoxNear, scanBoxFar = scanBoxMin, scanBoxMax
    elif atomCoord > scanBoxMax:
        scanBoxNear, scanBoxFar = scanBoxMax, scanBoxMin
    else:
        scanBoxNear = atomCoord
        # The farthest point is always a corner
        scanBoxFar = scanBoxMin if scanBoxMax-atomCoord < boxLen/2 else scanBoxMax
    return scanBoxNear, scanBoxFar


@njit(fastmath=True, cache=True)
def scanBox(minXYZ, scanBoxIdxs, scanBoxNearFarXYZs, boxLen,
            atomIdx, atomRad, atomXYZ, atomNeighIdxs,
            atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
            bufferDist=5.0, rmInSurf=True):
    """Find the nearest and furthest point of a given box from a given atom."""
    # Remove the box if it covers the inner surface
    if rmInSurf:
        scanBoxX = minXYZ[0] - bufferDist + (scanBoxIdxs[0]+1)*boxLen - boxLen*0.5
        scanBoxY = minXYZ[1] - bufferDist + (scanBoxIdxs[1]+1)*boxLen - boxLen*0.5
        scanBoxZ = minXYZ[2] - bufferDist + (scanBoxIdxs[2]+1)*boxLen - boxLen*0.5
        if not oppositeInnerAtoms(np.array((scanBoxX, scanBoxY, scanBoxZ)), atomXYZ, atomNeighIdxs,
                                  atomsSurfIdxs, atomsXYZ, atomsNeighIdxs):
            return 'none'

    # Check what does the box cover
    distNear = calcDist(atomXYZ, np.array(scanBoxNearFarXYZs[:3]))
    distFar = calcDist(atomXYZ, np.array(scanBoxNearFarXYZs[-3:]))
    if distNear < atomRad < distFar:
        return 'surf'
    elif distFar < atomRad:
        return 'bulk'
    else:
        return 'none'


# @annotate('scanAtom', color='cyan')
@njit(fastmath=True, cache=True)
def scanAtom(args):
    """Count the number of boxes that cover the outer spherical surface of a given atom."""
    magn, boxLen, minXYZ, atomIdx, atomRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs, bufferDist, rmInSurf = args
    atomXYZ, atomNeighIdxsPadded = atomsXYZ[atomIdx], atomsNeighIdxs[atomIdx]
    atomNeighIdxs = atomNeighIdxsPadded[atomNeighIdxsPadded > -1]

    atomX, atomY, atomZ = atomXYZ
    minX, minY, minZ = minXYZ
    atomBoxIdxX = int((atomX - minX + bufferDist)/boxLen)
    atomBoxIdxY = int((atomY - minY + bufferDist)/boxLen)
    atomBoxIdxZ = int((atomZ - minZ + bufferDist)/boxLen)
    numScan = ceil((atomRad + boxLen)/boxLen)
    atomSurfBoxs, atomBulkBoxs = [], []
    for i in range(-numScan, numScan + 1):
        scanBoxIdxX = atomBoxIdxX + i
        if scanBoxIdxX < 0 or scanBoxIdxX >= magn:
            continue
        scanBoxNearX, scanBoxFarX = getNearFarCoord(scanBoxIdxX, boxLen, minX, atomX, bufferDist)
        for j in range(-numScan, numScan + 1):
            scanBoxIdxY = atomBoxIdxY + j
            if scanBoxIdxY < 0 or scanBoxIdxY >= magn:
                continue
            scanBoxNearY, scanBoxFarY = getNearFarCoord(scanBoxIdxY, boxLen, minY, atomY, bufferDist)
            for k in range(-numScan, numScan + 1):
                scanBoxIdxZ = atomBoxIdxZ + k
                if scanBoxIdxZ < 0 or scanBoxIdxZ >= magn:
                    continue
                scanBoxNearZ, scanBoxFarZ = getNearFarCoord(scanBoxIdxZ, boxLen, minZ, atomZ, bufferDist)

                belong = scanBox(minXYZ, (scanBoxIdxX, scanBoxIdxY, scanBoxIdxZ),
                                 (scanBoxNearX, scanBoxNearY, scanBoxNearZ, scanBoxFarX, scanBoxFarY, scanBoxFarZ),
                                 boxLen,
                                 atomIdx, atomRad, atomXYZ, atomNeighIdxs,
                                 atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                                 bufferDist, rmInSurf)
                if belong == 'surf':
                    atomSurfBoxs.append((scanBoxIdxX, scanBoxIdxY, scanBoxIdxZ))
                elif belong == 'bulk':
                    atomBulkBoxs.append((scanBoxIdxX, scanBoxIdxY, scanBoxIdxZ))
    return atomSurfBoxs, atomBulkBoxs


# @annotate('scanAtomsForLoop', color='cyan')
@njit(fastmath=True, cache=True)
def scanAtomsForLoop(atomsIdxs, magn, boxLen, minXYZ,
                     atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                     bufferDist=5.0, rmInSurf=True):
    """Serialised loop to scan the atoms for timing comparison with the parallelised version."""
    allAtomsSurfBoxs, allAtomsBulkBoxs = [], []
    for atomIdx in atomsIdxs:
        scanAtomInp = (magn, boxLen, minXYZ, atomIdx,
                       atomsRad[atomIdx], atomsSurfIdxs, atomsXYZ, atomsNeighIdxs, 
                       bufferDist, rmInSurf)
        atomSurfBoxs, atomBulkBoxs = scanAtom(scanAtomInp)
        allAtomsSurfBoxs.extend(atomSurfBoxs)
        allAtomsBulkBoxs.extend(atomBulkBoxs)
    return allAtomsSurfBoxs, allAtomsBulkBoxs


# @annotate('scanAllAtoms', color='magenta')
def scanAllAtoms(args):
    """Count the number of boxes that cover the outer spherical surface of a set of atoms for a given box size."""
    magn, boxLen, atomsIdxs, minXYZ, atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs, bufferDist, rmInSurf, verbose, maxCPU = args
    scanAtomInps = [(magn, boxLen, minXYZ, atomIdx, atomsRad[atomIdx],
                     atomsSurfIdxs, atomsXYZ, atomsNeighIdxs, bufferDist, rmInSurf) for atomIdx in atomsIdxs]
    allAtomsSurfBoxs, allAtomsBulkBoxs = [], []
    with Pool(max_workers=maxCPU) as pool:
        for scanAtomResult in pool.map(scanAtom, scanAtomInps, chunksize=ceil(len(atomsIdxs) / maxCPU)):
            allAtomsSurfBoxs.extend(scanAtomResult[0])
            allAtomsBulkBoxs.extend(scanAtomResult[1])
    # allAtomsSurfBoxs, allAtomsBulkBoxs = scanAtomsForLoop(atomsIdxs, magn, boxLen, minXYZ,
    #                                                       atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
    #                                                       bufferDist, rmInSurf)
    allAtomsSurfBoxs, allAtomsBulkBoxs = set(allAtomsSurfBoxs), set(allAtomsBulkBoxs)
    allAtomsSurfBoxs.difference_update(allAtomsBulkBoxs)

    if verbose:
        epsInvStr = f"{1 / boxLen:.2f}"
        print(f"{epsInvStr.rjust(10)}{str(len(allAtomsBulkBoxs)).rjust(12)}{str(len(allAtomsSurfBoxs)).rjust(12)}")
    return allAtomsSurfBoxs, allAtomsBulkBoxs


# @annotate('writeBoxCoords', color='yellow')
def writeBoxCoords(atomsEle, atomsXYZ, allSurfBoxs, allBulkBoxs, minXYZ, boxLens, bufferDist,
                   outDir, npName):
    """Write out coordinates of scanned boxes."""
    minX, minY, minZ = minXYZ
    boxCoordsDir = f"{outDir}/boxCoords"
    if not isdir(boxCoordsDir):
        if not isdir(outDir):
            mkdir(outDir)
        mkdir(boxCoordsDir)
    with open(f"{boxCoordsDir}/{npName}_boxCoords.xyz", 'w') as f:
        for (i, boxLen) in enumerate(boxLens):
            if i != 0:
                f.write('\n')
            f.write(f"{len(atomsEle) + len(allSurfBoxs[i]) + len(allBulkBoxs[i])}\n")
            for (j, atomXYZ) in enumerate(atomsXYZ):
                f.write(f"\n{atomsEle[j]}\t{atomXYZ[0]} {atomXYZ[1]} {atomXYZ[2]}")
            for (boxIDX, boxIDY, boxIDZ) in allSurfBoxs[i]:
                boxX = minX - bufferDist + boxIDX*boxLen + boxLen/2
                boxY = minY - bufferDist + boxIDY*boxLen + boxLen/2
                boxZ = minZ - bufferDist + boxIDZ*boxLen + boxLen/2
                f.write(f"\nOV\t{boxX:.6f} {boxY:.6f} {boxZ:.6f}")
            for (boxIDX, boxIDY, boxIDZ) in allBulkBoxs[i]:
                boxX = minX - bufferDist + boxIDX*boxLen + boxLen/2
                boxY = minY - bufferDist + boxIDY*boxLen + boxLen/2
                boxZ = minZ - bufferDist + boxIDZ*boxLen + boxLen/2
                f.write(f"\nIV\t{boxX:.6f} {boxY:.6f} {boxZ:.6f}")


# @annotate('findAtomsWithSurfNeighs', color='cyan')
@njit(fastmath=True, cache=True)
def findAtomsWithSurfNeighs(atomsNeighIdxs, atomsSurfIdxs):
    """Find atoms with neighbours that are on the surface."""
    atomsIdxs = []
    for (atomIdx, atomNeighIdxs) in enumerate(atomsNeighIdxs):
        for neighIdx in atomNeighIdxs:
            if neighIdx < 0:
                break
            if neighIdx in atomsSurfIdxs:
                atomsIdxs.append(atomIdx)
                break
    return np.array(atomsIdxs) if len(atomsIdxs) > 0 else atomsSurfIdxs


# @annotate('exactBoxCnts', color='blue')
def exactBoxCnts(atomsEle, atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                 maxRange, minMaxBoxLens, minXYZ, npName, 
                 outDir='outputs', numCPUs=None, numBoxLen=10, bufferDist=5.0,
                 rmInSurf=True, writeBox=True, verbose=False):
    """
    Count the boxes that cover the outer surface of a set of overlapping spheres represented as exact spheres for different box sizes.
    
    Parameters
    ----------
    atomsEle : 1D ndarray of strs
        Element type of each atom.
    atomsRad : 1D ndarray of floats
        Radius of each atom.
    atomsSurfIdxs : 1D ndarray of ints
        Indices of surface atoms.
    atomsXYZ : 2D ndarray of floats
        Cartesian coordinates of each atom.
    atomsNeighIdxs : 2D ndarray of ints
        Neighbour atoms indices of each atom.
    maxRange : float
        Maximum range among all dimensions of the Cartesian space, defines the borders of the largest box.
    minMaxBoxLens : tuple of floats
        Minimum and maximum box lengths.
    minXYZ : 1D ndarray of floats
        Minimum values of each dimension in the Cartesian space.
    npName : str
        Identifier of the measured object, which forms part of the output file name, ideally unique.
    outDir : str, optional
        Path to the directory to store the output files.
    numCPUs : int, optional
        Number of CPUs to be used for parallelisation of tasks.
    numBoxLen : int, optional
        Number of box lengths to use for the collection of the box count data, spaced evenly on logarithmic scale.
    bufferDist : Union[int,float]
        Buffer distance from the borders of the largest box in Angstrom.
    rmInSurf : bool, optional
        Whether to remove the surface points on the inner surface.
    writeBox : bool, optional
        Whether to generate output files for visualisation.
    verbose : bool, optional
        Whether to display the details.
    
    Returns
    -------
    scales : list of floats
        Box lengths.
    counts : list of floats
        Number of boxes that cover the exact spherical surface of interest.
    
    Examples
    --------
    >>> eles, rads, xyzs, _, minxyz, maxxyz = readInp('example.xyz')
    >>> neighs, _ = findNN(rads, xyzs, minxyz, maxxyz, 1.2)
    >>> surfs = findSurf(xyzs, neighs, 'alphaShape', 5.0)
    >>> scalesES, countsES = exactBoxCnts(eles, rads, surfs, xyzs, neighs, 100, (0.2, 1), minxyz, 'example')
    """
    atomsIdxs = atomsSurfIdxs if rmInSurf else np.array(range(len(atomsEle)))
    if numCPUs is None: 
        numCPUs = cpu_count()
    # Resource allocations for parallelisation (current settings are based on empirical experiments -- optimised for the default minMaxBoxLens range), rooms available for further optimisation
    minAtomCPUperLen = max(1, len(atomsIdxs) // 25)
    maxBoxLenCPU = ceil(numBoxLen / 2)
    if numCPUs > maxBoxLenCPU * minAtomCPUperLen:
       atomConcMaxCPU = numCPUs // maxBoxLenCPU
       boxLenConcMaxCPU = maxBoxLenCPU
    elif numCPUs > minAtomCPUperLen:
        atomConcMaxCPU = minAtomCPUperLen
        boxLenConcMaxCPU = numCPUs // minAtomCPUperLen
    else:
        atomConcMaxCPU, boxLenConcMaxCPU = numCPUs, 1

    if verbose:
        print(f"  Representing the surface by treating each atom as exact spheres...")
        print(f"    Scanning over:\n      {numBoxLen} box lengths using {boxLenConcMaxCPU} cpu(s)...\n      {len(atomsIdxs)} atoms using {atomConcMaxCPU} cpu(s)...")
        print(f"    (1/eps)    (# bulk)    (# surf)")
    if writeBox:
        if not isdir(outDir):
            mkdir(outDir)

    overallBoxLen = maxRange + bufferDist * 2
    allLensSurfBoxs, allLensBulkBoxs, allLensSurfCnts, allLensBulkCnts = [], [], [], []
    scales, scanBoxLens, scanAllAtomsInps = [], [], []
    approxScanBoxLens = np.geomspace(minMaxBoxLens[1], minMaxBoxLens[0], num=numBoxLen)
    for approxScanBoxLen in approxScanBoxLens:  # Evenly reduced box lengths on log scale
        magnFac = int(overallBoxLen / approxScanBoxLen)
        scanBoxLen = overallBoxLen / magnFac
        scanAllAtomsInp = (magnFac, scanBoxLen, atomsIdxs, minXYZ,
                           atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs, bufferDist,
                           rmInSurf, verbose, atomConcMaxCPU)
        if boxLenConcMaxCPU > 1:
            scanAllAtomsInps.append(scanAllAtomsInp) 
        else:
            scanAllAtomsResult = scanAllAtoms(scanAllAtomsInp) 
            allAtomsSurfBoxs, allAtomsBulkBoxs = scanAllAtomsResult
            allLensSurfBoxs.append(allAtomsSurfBoxs)
            allLensBulkBoxs.append(allAtomsBulkBoxs)
            allLensSurfCnts.append(len(allAtomsSurfBoxs))

        scales.append(log10(magnFac / overallBoxLen))
        scanBoxLens.append(scanBoxLen)

    if boxLenConcMaxCPU > 1:
        with Pool(max_workers=atomConcMaxCPU) as pool:
            for scanAllAtomsResult in pool.map(scanAllAtoms, scanAllAtomsInps, 
                                               chunksize=ceil(numBoxLen / boxLenConcMaxCPU)):
                allAtomsSurfBoxs, allAtomsBulkBoxs = scanAllAtomsResult
                allLensSurfBoxs.append(allAtomsSurfBoxs)
                allLensBulkBoxs.append(allAtomsBulkBoxs)
                allLensSurfCnts.append(len(allAtomsSurfBoxs))
    counts = [log10(sCnt) if sCnt != 0 else np.nan for sCnt in allLensSurfCnts]

    if writeBox:
        writeBoxCoords(atomsEle, atomsXYZ, allLensSurfBoxs, allLensBulkBoxs, minXYZ, scanBoxLens, bufferDist, 
                       outDir, npName)
    return scales, counts




from math import ceil, floor, sqrt
from time import time

from numba import njit, prange
from numba.typed import List
import numpy as np
# from nvtx import annotate
from scipy.spatial import ConvexHull, Delaunay
from scipy.spatial._qhull import QhullError

from sphractal.constants import ATOMIC_RAD_DICT, METALLIC_RAD_DICT


def estDuration(func):
    """Return time taken to run a function."""
    def wrap(*arg, **kwargs):
        start = time()
        result = func(*arg, **kwargs)
        end = time()
        duration = end - start
        return result, duration
    return wrap


@njit(fastmath=True, cache=True)
def getMinMaxXYZ(atomsXYZ): 
    """Return the minimum and maximum values of each dimension of a set of coordinates."""
    rowNum, colNum = atomsXYZ.shape
    minXYZ, maxXYZ = np.empty(colNum), np.empty(colNum)
    for col in range(colNum):
        minVal = maxVal = atomsXYZ[0, col]
        for row in range(1, rowNum):
            val = atomsXYZ[row, col]
            minVal, maxVal = min(minVal, val), max(maxVal, val)
        minXYZ[col], maxXYZ[col] = minVal, maxVal
    return max(maxXYZ - minXYZ), minXYZ, maxXYZ


# @annotate('readInp', color='cyan')
def readInp(filePath, radType='atomic'):
    """Parse an xyz or a lmp file."""
    radDict = ATOMIC_RAD_DICT if radType == 'atomic' else METALLIC_RAD_DICT
    atomsEle, atomsRad, atomsXYZ = [], [], []
    numLinesSkip = 9 if '.lmp' in filePath else 2
    with open(filePath, 'r') as f:
        for (i, line) in enumerate(f):
            if i < numLinesSkip:
                continue
            eleXYZ = line.split()
            ele = eleXYZ[-4]
            atomsEle.append(ele)
            atomsRad.append(radDict[ele])
            atomX, atomY, atomZ = float(eleXYZ[-3]), float(eleXYZ[-2]), float(eleXYZ[-1])
            atomsXYZ.append((atomX, atomY, atomZ))
    atomsEle, atomsRad, atomsXYZ = np.array(atomsEle, dtype='U2'), np.array(atomsRad), np.array(atomsXYZ)
    maxRange, minXYZ, maxXYZ = getMinMaxXYZ(atomsXYZ)
    return atomsEle, atomsRad, atomsXYZ, maxRange, minXYZ, maxXYZ


@njit(fastmath=True, cache=True)
def allDirVecs():
    """Return a list of vectors corresponding to the standard 26 directions from a point."""
    dirVecs = []
    for x in range(-1, 2):
        for y in range(-1, 2):
            for z in range(-1, 2):
                dirVecs.append((x, y, z))
    return dirVecs


# @annotate('findNN', color='magenta')
@njit(fastmath=True, cache=True)
def findNN(atomsRad, atomsXYZ, minXYZ, maxXYZ, maxAtomRad, radMult=1.2, calcBL=False):
    """
    Compute the nearest neighbour list and average bond length for each atom.

    Parameters
    ----------
    atomsRad : 1D ndarray of floats
        Radius of each atom.
    atomsXYZ : 2D ndarray of floats
        Cartesian coordinates of each atom.
    minXYZ : 1D ndarray of floats
        Minimum values of each dimension in the Cartesian space.
    maxXYZ : 1D ndarray of floats
        Maximum values of each dimension in the Cartesian space.
    maxAtomRad : Union[int,float]
        Maximum value of atomic radius.
    radMult : Union[int,float]
        Multiplier to the atomic radii.
    calcBL : bool, optional
        Whether to compute the average bond length for each atom.

    Returns
    -------
    atomsNeighIdxsPadded : 2D ndarray of ints
        Neighbour atoms indices of each atom, padded with -1 for fixed number of columns.
    atomsAvgBondLen : 1D ndarray of floats
        Average bond lengths for each each.
    """
    (minX, minY, minZ), (maxX, maxY, maxZ) = minXYZ, maxXYZ
    atomsNeighIdxs = [[int(i) for i in range(0)] for _ in range(len(atomsRad))]
    atomsAvgBondLen = np.zeros_like(atomsRad)
    stepSize = maxAtomRad * 2 * radMult
    numX, numY, numZ = max(1, ceil((maxX-minX) / stepSize)), max(1, ceil((maxY-minY) / stepSize)), max(1, ceil((maxZ-minZ) / stepSize))
    boxes = [[[[int(i) for i in range(0)] for _ in range(numZ)] for _ in range(numY)] for _ in range(numX)]
    allDirections = allDirVecs()
    for (i, atom1rad) in enumerate(atomsRad):
        atom1X, atom1Y, atom1Z = atomsXYZ[i]

        x, y, z = floor((atom1X-minX) / stepSize), floor((atom1Y-minY) / stepSize), floor((atom1Z-minZ) / stepSize)
        for (dirX, dirY, dirZ) in allDirections:
            if 0 <= x + dirX < numX and 0 <= y + dirY < numY and 0 <= z + dirZ < numZ:
                for j in boxes[x + dirX][y + dirY][z + dirZ]:
                    atom2X, atom2Y, atom2Z = atomsXYZ[j]
                    atom2rad = atomsRad[j]

                    diffX, diffY, diffZ = abs(atom1X - atom2X), abs(atom1Y - atom2Y), abs(atom1Z - atom2Z)
                    sumOfSquares = diffX*diffX + diffY*diffY + diffZ*diffZ
                    if sumOfSquares < ((atom1rad+atom2rad)*radMult) ** 2:
                        atomsNeighIdxs[i].append(j)
                        atomsNeighIdxs[j].append(i)
                        if calcBL:
                            atomsAvgBondLen[i] += sqrt(sumOfSquares)
                            atomsAvgBondLen[j] += sqrt(sumOfSquares)
        boxes[x][y][z].append(i)

    # Turn list of lists into NumPy array padded with -1
    maxNeighNum = max([len(atomNeighIdxs) for atomNeighIdxs in atomsNeighIdxs])
    atomsNeighIdxsPadded = np.full((len(atomsNeighIdxs), maxNeighNum), -1)
    for (i, atomNeighIdx) in enumerate(atomsNeighIdxs):
        atomsNeighIdxsPadded[i][:len(atomNeighIdx)] = atomNeighIdx

    if calcBL:
        for atomID in range(len(atomsRad)):
            atomsAvgBondLen[atomID] /= len(atomsNeighIdxs[atomID])
    return atomsNeighIdxsPadded, atomsAvgBondLen


@njit(fastmath=True, cache=True)
def findTetras(tetraVtxsIdxs, r, alpha):
    """Return tetrahedrons with their circumsphere radii smaller than a specified alpha value."""
    return tetraVtxsIdxs[r < alpha, :]


@njit(fastmath=True, cache=True)
def rmDupTris(tris):
    """Remove triangles that occur twice (internal triangles)."""
    unqTris = []
    for tri in tris:
        if tris.count(tri) == 1:
            unqTris.append(tri)
    return unqTris


# @annotate('alphaShape', color='cyan')
def alphaShape(atomsXYZ, tetraVtxsIdxs, alpha):
    """
    Return points that form the surface using alpha shape algorithm.

    Algorithm modified from https://stackoverflow.com/questions/26303878/alpha-shapes-in-3d
    Radius of the sphere fitting inside the tetrahedral < alpha (http://mathworld.wolfram.com/Circumsphere.html)
    """
    # Find radius of the circumsphere
    tetraVtxs = atomsXYZ[tetraVtxsIdxs]
    norm2 = np.sum(tetraVtxs ** 2, axis=2)[:, :, np.newaxis]
    ones = np.ones((tetraVtxs.shape[0], tetraVtxs.shape[1], 1))
    a = np.linalg.det(np.concatenate((tetraVtxs, ones), axis=2))
    Dx = np.linalg.det(np.concatenate((norm2, tetraVtxs[:, :, [1, 2]], ones), axis=2))
    Dy = np.linalg.det(np.concatenate((norm2, tetraVtxs[:, :, [0, 2]], ones), axis=2))
    Dz = np.linalg.det(np.concatenate((norm2, tetraVtxs[:, :, [0, 1]], ones), axis=2))
    c = np.linalg.det(np.concatenate((norm2, tetraVtxs), axis=2))
    with np.errstate(divide='ignore', invalid='ignore'):
        r = np.sqrt(Dx**2 + Dy**2 + Dz**2 - 4*a*c) / (2*np.abs(a))

    # Find tetrahedrons and triangles
    tetras = findTetras(tetraVtxsIdxs, r, alpha)
    triComb = np.array([(0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3)])
    tris = tetras[:, triComb].reshape(-1, 3)
    tris = rmDupTris(List(tuple(i) for i in tris))
    return np.unique(np.concatenate(tris))


# @annotate('findSurf', color='yellow')
def findSurf(atomsXYZ, atomsNeighIdxs, option='alphaShape', alpha=3.0, bulkCN=12):
    """
    Return the indices of surface atoms.
    
    Parameters
    ----------
    atomsXYZ : 2D ndarray of floats
        Cartesian coordinates of each atom.
    atomsNeighIdxs : 2D ndarray of ints
        Neighbour atoms indices of each atom.
    option : {'alphaShape', 'convexHull', 'numNeigh'}, optional
        Algorithm to identify the spheres on the surface. 
        'convexHull' tends to identify less surface atoms; 'numNeigh' tends to identify more surface atoms.
        'alphaShape' is a generalisation of 'convexHull'.
    alpha : Union[int, float], optional
        'alpha' for the alpha shape algorithm, only used if 'option' is 'alphaShape'.
    bulkCN : int, optional
        Minimum number of neighbouring atoms for a non-surface atom.
    
    Returns
    -------
    atomsSurfIdxs : 1D ndarray of ints
        Indices of surface atoms.
    
    Notes
    -----
    Other alternatives include:
    - https://dl.acm.org/doi/abs/10.1145/2073304.2073339
    - https://onlinelibrary.wiley.com/doi/pdf/10.1002/jcc.25384
    - https://www.jstage.jst.go.jp/article/tmrsj/45/4/45_115/_article
    """
    atomsSurfIdxs = np.zeros(len(atomsXYZ), dtype=np.bool_)
    if option == 'convexHull':
        for atomIdx in np.array(ConvexHull(atomsXYZ).vertices):
            atomsSurfIdxs[atomIdx] = True
    elif option == 'numNeigh':
        for (atomIdx, atomNeighIdxs) in enumerate(atomsNeighIdxs):
            if len(atomNeighIdxs[atomNeighIdxs > -1]) < bulkCN:
                atomsSurfIdxs[atomIdx] = True
    elif option == 'alphaShape':
        try:
            tetraVtxsIdxs = Delaunay(atomsXYZ).simplices
            for atomIdx in alphaShape(atomsXYZ, tetraVtxsIdxs, alpha):
                atomsSurfIdxs[atomIdx] = True
        except QhullError:
            atomsSurfIdxs = np.full(len(atomsXYZ), True)
    atomsSurfIdxs = np.where(atomsSurfIdxs)[0]
    return atomsSurfIdxs


@njit(fastmath=True, cache=True)
def findAtomNeighs(neighIdxs, atomsSurfIdxs):
    """Divide a list of atomic neighbours into two based on whether they lie on the surface."""
    surfNeighIdxs, bulkNeighIdxs = [], []
    for idx in neighIdxs:
        if idx in atomsSurfIdxs:
            surfNeighIdxs.append(idx)
        else:
            bulkNeighIdxs.append(idx)
    return np.array(surfNeighIdxs, dtype=np.uint32), np.array(bulkNeighIdxs, dtype=np.uint32)


@njit(fastmath=True, cache=True)  # parallel=True slows things down, incompatible with multiprocessing.Pool()
def calcDist(p1, p2):
    """Return Euclidean distance between two points."""
    d = 0
    for i in prange(3):
        d += (p2[i] - p1[i]) ** 2
    return d ** 0.5


@njit(fastmath=True, cache=True)
def getOrdSurfNeighCombs(surfNeighDists, surfNeighIdxs):
    """Compute ID pairs of a set of neighbouring atoms, ordered by distance from a given point."""
    ordSurfNeighIdxs = surfNeighIdxs[np.argsort(surfNeighDists)]
    surfNeighIdxPairs = np.stack(np.triu_indices(len(ordSurfNeighIdxs), k=1), axis=-1)
    return [ordSurfNeighIdxs[idxPair] for idxPair in surfNeighIdxPairs]


@njit(fastmath=True, cache=True)
def closestSurfAtoms(pointXYZ, surfNeighIdxs, atomsXYZ, atomsNeighIdxs):
    """Return two closest surface atom pairs from a given point, that are neighbour of each other."""
    if len(surfNeighIdxs) < 2: 
        return np.array([np.nan]), np.array([np.nan])
    surfNeighXYZs = atomsXYZ[surfNeighIdxs]
    surfNeighDists = np.array([calcDist(pointXYZ, surfNeighXYZ) for surfNeighXYZ in surfNeighXYZs])
    for idxPair in getOrdSurfNeighCombs(surfNeighDists, surfNeighIdxs):
        if idxPair[0] in atomsNeighIdxs[idxPair[1]]:
            return atomsXYZ[idxPair[0]], atomsXYZ[idxPair[1]]
    return np.array([np.nan]), np.array([np.nan])


@njit(fastmath=True, cache=True)
def oppositeInnerAtoms(pointXYZ, atom1XYZ, atomNeighIdxs, 
                       atomsSurfIdxs, atomsXYZ, atomsNeighIdxs):
    """Return whether a point lies opposite of the average coordinates of neighbouring inner atoms of a given atom."""
    surfNeighIdxs, bulkNeighIdxs = findAtomNeighs(atomNeighIdxs, atomsSurfIdxs)
    atom2XYZ, atom3XYZ = closestSurfAtoms(pointXYZ, surfNeighIdxs, atomsXYZ, atomsNeighIdxs)
    innerNeighXYZs = atomsXYZ[bulkNeighIdxs] if len(bulkNeighIdxs) > 0 else atomsXYZ[surfNeighIdxs]
    avgInnerAtomXYZ = np.array([innerNeighXYZs[:, i].mean() for i in range(3)])
    normal = avgInnerAtomXYZ - atom1XYZ if len(atom3XYZ) == 1 else np.cross(atom2XYZ - atom1XYZ, atom3XYZ - atom1XYZ)
    # If the given point is on the opposite side of the surface plane, the sign of products of the dot products will be negative
    return np.dot(normal, avgInnerAtomXYZ - atom1XYZ) * np.dot(normal, pointXYZ - atom1XYZ) < 0



from os import mkdir
from os.path import isdir
from warnings import warn

import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
import numpy as np
from statsmodels.api import OLS, add_constant

from sphractal.constants import PLT_PARAMS
from sphractal.utils import findNN, findSurf, readInp
from sphractal.surfVoxel import voxelBoxCnts
from sphractal.surfExact import exactBoxCnts
# from sphractal.utils import estDuration, annotate


# @annotate('findSlope', color='green')
def findSlope(scales, counts, npName='', outDir='outputs', trimLen=True,
              minSample=6, confLvl=95, 
              visReg=True, figType='paper', saveFig=False, showPlot=False, verbose=False):
    """
    Compute the slope (box counting dimension) from the box-counting data collected.

    Parameters
    ----------
    scales : list of floats
        Box lengths.
    counts : list of floats
        Number of boxes that cover the exact spherical surface of interest.
    npName : str, optional
        Identifier of the measured object, which forms part of the output file name, ideally unique.
    outDir : str, optional
        Path to the directory to store the output files.
    trimLen : bool, optional
        Whether to remove the box counts obtained using boxes of extreme sizes.
    minSample : int, optional
        Minimum number of box count data points to be retained for slope estimation from the linear regression fitting.
    confLvl : Union[int, float]
        Confidence level of confidence interval in percentage.
    visReg : bool, optional
        Whether to generate figures from the linear regression fitting process.
    figType : {'paper', 'poster', 'talk'}
        Type of figures to be generated.
    saveFig : bool, optional
        Whether to save the final figure generated, only works when 'visReg' is True.
    showPlot : bool, optional
        Whether to show the figures generated, only works when 'visReg' is True.
    verbose : bool, optional
        Whether to display the details.
    
    Returns
    -------
    r2score : float
        Coefficient of determination from determination of the dimension of point clouds surface.
    boxCntDim : float
        Box-counting dimension of the point clouds representation of the surface.
    slopeCI : 1D ndarray of floats
        Confidence interval of the box-counting dimension of the point clouds surface.
    minMaxLens : tuple of floats
        Minimum and maximum box lengths used to determine slope.
    """
    assert isinstance(minSample, int), 'minSample should be integer!'

    if visReg:
        plt.rc('font', family='sans-serif')
        plt.rc('xtick', labelsize='x-small')
        plt.rc('ytick', labelsize='x-small')
        params = PLT_PARAMS[figType]
        figSize, dpi, fontSize, labelSize = params['figSize'], params['dpi'], params['fontSize'], params['labelSize']
        legendSize, lineWidth, markerSize = params['legendSize'], params['lineWidth'], params['markerSize']
    else:
        figSize = dpi = fontSize = labelSize = legendSize = lineWidth = markerSize = None

    # Remove invalid entries in the box counts data collected
    while np.nan in counts:
        nanIdx = counts.index(np.nan)
        del counts[nanIdx]
        del scales[nanIdx]

    if abs(confLvl) > 100:
        warn(f"Confidence level out of range, confidence intervals are unreliable! 'confLvl' should be within [0, 100) "
             f"instead of {confLvl}")
    alphaCI = 1 - confLvl/100

    firstPointIdx, lastPointIdx, removeSmallBoxes = 0, len(scales), True
    r2score, boxCntDim, slopeCI, minMaxLens = 0.0, 0.0, np.array((np.inf, np.inf)), (scales[0], scales[-1])
    r2scorePrev, boxCntDimPrev, slopeCIPrev, minMaxLensPrev = 0.0, 0.0, np.array((np.inf, np.inf)), (scales[0], scales[-1])
    while len(scales[firstPointIdx:lastPointIdx]) >= minSample:

        x, y = scales[firstPointIdx:lastPointIdx], counts[firstPointIdx:lastPointIdx]
        regModel = OLS(endog=y, exog=add_constant(x)).fit()
        r2score, boxCntDim, slopeCI = regModel.rsquared, regModel.params[1], regModel.conf_int(alpha=alphaCI)[1]
        minMaxLens = (x[0], x[-1])
        yPred = regModel.predict()  # Returns ndarray, allowing subtraction later

        if visReg:
            plt.close()
            fig = plt.figure(figsize=figSize, dpi=dpi)
            ax = fig.add_subplot(1, 1, 1)
            handleScatter = ax.scatter(x, y, marker='o', s=markerSize, c='r', alpha=1, edgecolors='k', linewidths=1.2,
                                       zorder=3)
            handleBestFit = ax.plot(x, yPred, linestyle='-', linewidth=1., color='k', label='OLS')
            ax.grid(linestyle='dotted')

            # Compute confidence bands
            predOLS = regModel.get_prediction()
            lowCIs, upCIs = predOLS.summary_frame()['mean_ci_lower'], predOLS.summary_frame()['mean_ci_upper']
            handleConfBand = ax.plot(x, upCIs, linestyle='--', linewidth=lineWidth, color='b')
            ax.plot(x, lowCIs, linestyle='--', linewidth=lineWidth, color='b')
            ax.fill_between(x, upCIs, lowCIs, alpha=0.2)

            ax.set_xlabel(r'log$(1/\epsilon)$', fontsize=labelSize)
            ax.set_ylabel(r'log$(N)$', fontsize=labelSize)
            ax.yaxis.set_major_formatter(FormatStrFormatter('% 1.1f'))
            # ax.set_title('', fontsize=fontSize)
            ax.legend(handles=(handleScatter, handleBestFit[0], handleConfBand[0]), 
                      labels=('Actual box counts', fr"Best fit line ($R^2$: {r2score:.3f})",
                              f"{confLvl}% confidence bands"),
                      title=fr"$D_{{box}}$: {boxCntDim:.3f} [{slopeCI[0]:.3f}, {slopeCI[1]:.3f}]",
                      title_fontsize=legendSize,
                      fontsize=legendSize)
            plt.tight_layout()

        # Removal of next point (beware of weird behaviour in middle range)
        # lstSqErrs = np.subtract(y, yPred) ** 2
        # if len(y) % 2 == 0:
        #     lowBoundErrSum, upBoundErrSum = lstSqErrs[:len(y) // 2].sum(), lstSqErrs[len(y) // 2:].sum()
        # else:
        #     lowBoundErrSum, upBoundErrSum = lstSqErrs[:len(y) // 2].sum(), lstSqErrs[len(y) // 2 + 1:].sum()
        # if lowBoundErrSum > upBoundErrSum: firstPointIdx += 1
        # else: lastPointIdx -= 1

        if saveFig:
            boxCntDimsDir = f"{outDir}/boxCntDims"
            if not isdir(boxCntDimsDir):
                if not isdir(outDir):
                    mkdir(outDir)
                mkdir(boxCntDimsDir)
            plt.savefig(f"{boxCntDimsDir}/{npName}_boxCntDim.png", bbox_inches='tight')
        if showPlot:
            plt.show()

        if trimLen:
            if removeSmallBoxes:
                if round(r2score, 3) < round(r2scorePrev, 3):
                    removeSmallBoxes = False
                lastPointIdx -= 1
            else:
                if round(r2score, 3) < round(r2scorePrev, 3):
                    if verbose:
                        print(f"  D_Box: {boxCntDimPrev:.4f} [{slopeCIPrev[0]:.4f}, {slopeCIPrev[1]:.4f}],  R2: {r2scorePrev:.4f},  boxLens: ({minMaxLensPrev[0]:.4f}, {minMaxLensPrev[1]:.4f})\n")
                    return r2scorePrev, boxCntDimPrev, slopeCIPrev, minMaxLensPrev
                firstPointIdx += 1
        else:
            break
        r2scorePrev, boxCntDimPrev, slopeCIPrev, minMaxLensPrev = r2score, boxCntDim, slopeCI, minMaxLens

    if verbose:
        print(f"  D_Box: {boxCntDim:.4f} [{slopeCI[0]:.4f}, {slopeCI[1]:.4f}],  R2: {r2score:.4f},  boxLens: ({minMaxLens[0]:.4f}, {minMaxLens[1]:.4f})\n")
    return r2score, boxCntDim, slopeCI, minMaxLens


# @annotate('runCase', color='cyan')
# @estDuration
def runBoxCnt(inpFilePath, 
              radType='atomic', radMult=1.2, calcBL=False, findSurfAlg='alphaShape', alphaMult=2.0, bulkCN=12,
              outDir='outputs', trimLen=True, minSample=6, confLvl=95, 
              rmInSurf=True, vis=True, figType='paper', saveFig=False, showPlot=False, verbose=False,
              voxelSurf=True, numPoints=10000, gridNum=1024, exePath='$FASTBC', genPCD=False,
              exactSurf=True, minLenMult=0.25, maxLenMult=1, numCPUs=8, numBoxLen=10, bufferDist=5.0, writeBox=True): 
    """
    Run box-counting algorithm on the surface of a given atomistic object consisting of a set of spheres represented as either a voxelised point cloud or mathematically precise object.
    
    Parameters
    ----------
    inpFilePath : str
        Path to xyz file containing Cartesian coordinates of a set of atoms.
    radType : {'atomic', 'metallic'}, optional
        Type of radii to use for the atoms.
    radMult : Union[int, float], optional
        Multiplier to the radii of atoms to identify their neighbouring atoms.
    calcBL : bool, optional
        Whether to compute the average distance from its neighbours for each atom.
    findSurfAlg : {'alphaShape', 'convexHull', 'numNeigh'}, optional
        Algorithm to identify the surface atoms.
    alphaMult : Union[int, float], optional
        Multiplier to the minimum radius to decide 'alpha' value for the alpha shape algorithm, only used if
        'findSurfAlg' is 'alphaShape'. Recommendation: 
        2.0 * 100% ATOMIC_RAD == 5/3 * 120% ATOMIC_RAD ~= 100% METALLIC_RAD * 2.5
    bulkCN : int, optional
        Minimum number of neighbouring atoms for non-surface atoms.
    outDir : str, optional
        Path to directory to store the output files.
    trimLen : bool, optional
        Whether to remove the box counts obtained using boxes of extreme sizes.
    minSample : int, optional
        Minimum number of data points to retain for slope estimation from the linear regression fitting.
    confLvl : Union[int, float], optional
        Confidence level of confidence intervals (%).
    rmInSurf : bool, optional
        Whether to remove the inner surface points.
    vis : bool, optional
        Whether to generate files for visualisation.
    figType : {'paper', 'poster', 'talk', 'notebook'}
        Research purpose of the figures generated.
    saveFig : bool, optional
        Whether to save the figures generated from linear regression fitting, only used if 'vis' is True.
    showPlot : bool, optional
        Whether to show the figures generated from linear regression fitting, only used if 'vis' is True.
    verbose : bool, optional
        Whether to display the details.
    voxelSurf : bool, optional
        Whether to represent the surface as voxelised point clouds.
    numPoints : int, optional
        Number of surface points to generate around each atom.
    gridNum : int, optional
        Resolution of the 3D binary image.
    exePath : str, optional
        Path to the compiled executable of the C++ code for box-counting on 3D binary image written by Ruiz de Miras et al.
    genPCD : bool, optional
        Whether to generate point cloud data (pcd) file.
    exactSurf : bool, optional
        Whether to represent the surface in a mathematically exact manner.
    minLenMult : float, optional
        Multiplier to the minimum radius to determine the minimum box length.
    maxLenMult : float, optional
        Multiplier to the maximum radius to determine the maximum box length.
    numCPUs : int, optional
        Number of cores to be used for parallelisation.
    numBoxLen : int, optional
        Number of box length samples for the collection of the box count data, spaced evenly on logarithmic scale.
    bufferDist : Union[int,float]
        Buffer distance from the borders of the largest box (Angstrom).
    writeBox : bool, optional
        Whether to generate output file containing coordinates of examined boxes.
    
    Returns
    -------
    r2VX : float
        Coefficient of determination from linear regression fitting to the box-counting data obtained from the voxelised point cloud surface representation.
    bcDimVX : float
        Box-counting dimension of the voxelised point cloud representation of the surface.
    confIntVX : 1D ndarray of floats
        Confidence interval of the slope estimated for the voxelised point cloud surface representation.
    minMaxLensVX : tuple of floats
        Minimum and maximum box lengths for the final slope determination for the voxelised point cloud surface representation.
    r2EX : float
        Coefficient of determination from linear regression fitting to the box-counting data obtained from the mathematically exact surface representation.
    bcDimEX : float
        Box-counting dimension of the mathematically exact representation of the surface.
    confIntEX : 1D ndarray of floats
        Confidence interval of the slope estimated for the mathematically exact surface representation.
    minMaxLensEX : tuple of floats
        Minimum and maximum box lengths for final slope determination for the mathematically exact surface representation.
    
    Examples
    --------
    >>> r2VX, bcDimVX, confIntVX, minMaxLensVX, r2EX, bcDimEX, confIntEX, minMaxLensEX = runBoxCnt('example.xyz')
    """
    atomsEle, atomsRad, atomsXYZ, maxRange, minXYZ, maxXYZ = readInp(inpFilePath, radType)
    atomsNeighIdxs, atomsAvgBondLen = findNN(atomsRad, atomsXYZ, minXYZ, maxXYZ, atomsRad.max(), radMult, calcBL)
    atomsSurfIdxs = findSurf(atomsXYZ, atomsNeighIdxs, findSurfAlg, alphaMult * atomsRad.min(), bulkCN)
    testCase = inpFilePath.split('/')[-1][:-4]
    if verbose:
        print(f"\n{testCase}")

    r2VX, bcDimVX, confIntVX, minMaxLensVX = np.nan, np.nan, (np.nan, np.nan), (np.nan, np.nan)
    r2EX, bcDimEX, confIntEX, minMaxLensEX = np.nan, np.nan, (np.nan, np.nan), (np.nan, np.nan)
    if not isdir(outDir):
        mkdir(outDir)
    if voxelSurf:
        scalesVX, countsVX = voxelBoxCnts(atomsEle, atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                                          testCase, outDir, numCPUs, exePath,
                                          radType, numPoints, gridNum,
                                          rmInSurf, vis, verbose, genPCD)
        r2VX, bcDimVX, confIntVX, minMaxLensVX = findSlope(scalesVX, countsVX, f"{testCase}_VX", outDir, trimLen,
                                                           minSample, confLvl, vis, figType, saveFig, showPlot, verbose)
    if exactSurf:
        minAtomRad = atomsRad.min()
        scalesEX, countsEX = exactBoxCnts(atomsEle, atomsRad, atomsSurfIdxs, atomsXYZ, atomsNeighIdxs,
                                          maxRange, (minAtomRad * minLenMult, minAtomRad * maxLenMult),
                                          minXYZ, testCase, outDir, numCPUs, numBoxLen, bufferDist,
                                          rmInSurf, writeBox, verbose)
        r2EX, bcDimEX, confIntEX, minMaxLensEX = findSlope(scalesEX, countsEX, f"{testCase}_EX", outDir, trimLen,
                                                           minSample, confLvl, vis, figType, saveFig, showPlot, verbose)
    return r2VX, bcDimVX, confIntVX, minMaxLensVX, r2EX, bcDimEX, confIntEX, minMaxLensEX
